package Mechanism;
#
#    Mechanism.pm: perl module to assist with reading, interpreting,
#                  and writing chemical mechanism files
#    Copyright (C) 2017  Tim Butler <tim.butler@iass-potsdam.de>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
use strict;
use warnings;
use Carp;
#use KPP::Family;

# Author: Tim Butler
# Amended 10/01/13: Jane Coates
# 4/4/13: Tim Butler: fixed double counting bug in family calculation

sub new {
	my $class = shift;
	my $self = {};
	bless $self, $class;
}

sub read_KPP {
	my ($self, $filename) = @_;
	my (@original, %reactions, %produced, %consumed, %reactant_string, %reaction_string);
	my %full_string;
	my %rate_string;
	my %reaction_number;
	croak "$filename: No such file" unless (-f $filename);
	open FILE, $filename or croak $!;
	$original[0] = "There is no reaction zero\n";
	my $rnumber = 0;
	foreach my $line (<FILE>)
	{
		my ($label, $reactants, $products, $rate, $extra)
			= $line =~ m/
				^\s* {\#(.*?)}      # reaction label
				(.*?) = (.*?)       # reactants = products
				: \s*               # reaction labels (removed, now part of the rate)
				(.*?);              # rate expression terminated by ";"
				(.*?)$              # everything else
			/x;

		next unless (defined $label and defined $reactants and defined $products and defined $rate);
        croak "Multiple definition of reaction $label\n" if (defined $reaction_number{$label});

		push @original, $line;
		$reaction_number{$label} = $rnumber++;
		$full_string{$label} = $line;
		$reactants =~ s/{.*?}//g;
		$products =~ s/{.*?}//g;
		$reactants =~ s/^\s*(.*?)\s*$/$1/;
		$products =~ s/^\s*(.*?)\s*$/$1/;

		if ($products =~ /-/) {
			carp "Products include subtractions ($products)";
		}

		my @reactants = split /\s*\+\s*/, $reactants;
		my @products = split /\s*\+\s*/, $products;

		$reactants =~ s/\s+/ /g;
		$reactant_string{$label} = $reactants;
		$reaction_string{$label} = "$reactants = $products";
		$rate_string{$label} = $rate;

		my %seen = ();
		foreach (@reactants) {
			my ($fraction, $species) = &_get_fraction($_);
			if (defined $seen{$species}) {
				$reactions{$label}{reactants}{$species} += $fraction;
			} else  {
				push @{ $consumed{$species}{reactions} }, $label;
				$reactions{$label}{reactants}{$species} = $fraction;
				$seen{$species} = 1;
			}
		}

		%seen = ();
		foreach (@products) {
			my ($fraction, $species) = &_get_fraction($_);
			if (defined $seen{$species}) {
				$reactions{$label}{products}{$species} += $fraction;
			} else  {
				push @{ $produced{$species}{reactions} }, $label;
				$reactions{$label}{products}{$species} = $fraction;
				$seen{$species} = 1;
			}
		}
    }
	close FILE;

	$self->{reactions} = \%reactions;
	$self->{consumed} = \%consumed;
	$self->{produced} = \%produced;
	$self->{original} = \@original;
	$self->{reaction_number} = \%reaction_number;
	$self->{reactant_string} = \%reactant_string;
	$self->{reaction_string} = \%reaction_string;
	$self->{rate_string} = \%rate_string;
	$self->{full_string} = \%full_string;

	return $self;
}

sub read_MOZART {
	my ($self, $filename) = @_;
	my (@original, %reactions, %produced, %consumed, %reactant_string, %reaction_string);
	my %full_string;
	my %rate_string;
	my %reaction_number;
	croak "$filename: No such file" unless (-f $filename);
	open FILE, $filename or croak $!;
	my $text = join '', <FILE>;
	close FILE;
	my ($photolysis) = $text =~ /Photolysis(.*?)End Photolysis/s;
	my ($thermal)    = $text =~ /Reactions(.*?)End Reactions/s;
	$photolysis = "" unless defined $photolysis;
	$text = $photolysis . $thermal;
	$original[0] = "There is no reaction zero\n";
	my $rnumber = 0;
	$text =~ s/^\s+$//mg;
	$text =~ s/^\*.*$//mg;
	$text =~ s/\n+/\n/sg;
	$text =~ s/^\s*(.*?)\s*$/$1/s;
	my @lines = split /\n/, $text;
	for (my $l = 0; $l < @lines; $l++) {
		my $line = $lines[$l];
		my ($usr_rate, $rxntxt, $simple_rate);
		if ($line =~ s/(\[.*?\])//) {
			$usr_rate = $1;
		} else {
			$usr_rate = '';
		}
		if ($line =~ s/;\s*(.*?)\s*$//) {
			$simple_rate = $1;
		} else {
			$simple_rate = '';
		}
		$rxntxt = $line;
		while ($l < $#lines and $lines[$l+1] =~ /^\s*\+/) { # concatenate multi-line reaction rates
			$rxntxt .= $lines[++$l];
		}
		$rxntxt =~ s/\*/ /g;
		$rxntxt =~ s/\s+/ /g;
		$rxntxt =~ s/^\s*(.*?)\s*$/$1/;
		my ($reactants, $products) = split ' -> ', $rxntxt;
		$products = "" unless defined $products;

		push @original, $line;
		my $label = $rnumber++;
		$reaction_number{$label} = $rnumber;
		$full_string{$label} = "$usr_rate $rxntxt ; $simple_rate\n";
		$reactants =~ s/{.*?}//g;
		$products =~ s/{.*?}//g;
		$reactants =~ s/^\s*(.*?)\s*$/$1/;
		$products =~ s/^\s*(.*?)\s*$/$1/;

		if ($products =~ /-/) {
			carp "Products include subtractions ($products)";
		}

		my @reactants = split /\s*\+\s*/, $reactants;
		my @products = split /\s*\+\s*/, $products;

		$reactants =~ s/\s+/ /g;
		$reactant_string{$label} = $reactants;
		$reaction_string{$label} = "$reactants = $products";
		$rate_string{$label} = "$usr_rate ; $simple_rate";

		my %seen = ();
		foreach (@reactants) {
			my ($fraction, $species) = &_get_fraction($_);
			if (defined $seen{$species}) {
				$reactions{$label}{reactants}{$species} += $fraction;
			} else  {
				push @{ $consumed{$species}{reactions} }, $label;
				$reactions{$label}{reactants}{$species} = $fraction;
				$seen{$species} = 1;
			}
		}

		%seen = ();
		foreach (@products) {
			my ($fraction, $species) = &_get_fraction($_);
			if (defined $seen{$species}) {
				$reactions{$label}{products}{$species} += $fraction;
			} else  {
				push @{ $produced{$species}{reactions} }, $label;
				$reactions{$label}{products}{$species} = $fraction;
				$seen{$species} = 1;
			}
		}
    }

	$self->{reactions} = \%reactions;
	$self->{consumed} = \%consumed;
	$self->{produced} = \%produced;
	$self->{original} = \@original;
	$self->{reaction_number} = \%reaction_number;
	$self->{reactant_string} = \%reactant_string;
	$self->{reaction_string} = \%reaction_string;
	$self->{rate_string} = \%rate_string;
	$self->{full_string} = \%full_string;

	return $self;
}


sub _get_fraction {
	my $string = shift;
	my ($fraction, $species);
	unless (($fraction, $species) = $string =~ /^([.\d]+)\s*(.*)/) {
		($fraction, $species) = (1, $string);
	}
	return ($fraction, $species);
}

# Return a list of all reactions producing a particular product, and
# another list with the yields of that product from each of those reactions
sub producing {
	my $self = shift;
	my $species = $_[0];
	my (@reactions);
	foreach my $reaction (@{ $self->{produced}{$species}{reactions} }) {
		push @reactions, $reaction; 
	}
	return \@reactions; 
}

sub producing_from {
	my $self = shift;
	my ($species, $from) = @_;
	my (@reactions); 
	foreach my $reaction (@{ $self->{produced}{$species}{reactions} }) {
		next unless exists $self->{reactions}{$reaction}{reactants}{$from};
		push @reactions, $reaction; 
	}
	return \@reactions; 
}

sub consuming {
	my $self = shift;
	my $species = shift;
	my (@reactions); 
	foreach my $reaction (@{ $self->{consumed}{$species}{reactions} }) {
		push @reactions, $reaction; 
	}
	return \@reactions; 
}

sub reacting_with {
	my $self = shift;
	my ($reactant1, $reactant2) = @_;
	my (@reactions); 
	foreach my $reaction ( keys %{ $self->{reactions} }) {
		next unless exists $self->{reactions}{$reaction}{reactants}{$reactant1}
			and exists $self->{reactions}{$reaction}{reactants}{$reactant2};
		push @reactions, $reaction;
	}
	return \@reactions; 
}

sub yield_of {
	my $self = shift;
	my ($species, $reactions) = @_;
	my @yields;
	foreach my $reaction (@$reactions) {
		if (exists $self->{reactions}{$reaction}{products}{$species}) {
			push @yields, $self->{reactions}{$reaction}{products}{$species};
		}
		else {
			push @yields, 0;
		}
	}
	return \@yields;
}

sub effect_on {
	my $self = shift;
	my ($species, $reactions) = @_;
	my @effects;
	foreach my $reaction (@$reactions) {
		my $effect = 0;
		if (exists $self->{reactions}{$reaction}{products}{$species}) {
			$effect += $self->{reactions}{$reaction}{products}{$species};
#print "Product contrib $self->{reactions}{$reaction}{products}{$species}\n" if $species eq 'OH' and $reaction eq '127';
		} elsif (exists $self->{reactions}{$reaction}{family_products}{$species}) {
			$effect += $self->{reactions}{$reaction}{family_products}{$species};
		}
		if (exists $self->{reactions}{$reaction}{reactants}{$species}) {
			$effect -= $self->{reactions}{$reaction}{reactants}{$species};
#print "Reactant contrib $self->{reactions}{$reaction}{reactants}{$species}\n" if $species eq 'OH' and $reaction eq '127';
		} elsif (exists $self->{reactions}{$reaction}{family_reactants}{$species}) {
			$effect -= $self->{reactions}{$reaction}{family_reactants}{$species};
		}
		#print "Foo: $reaction $species $effect\n";
		push @effects, $effect;
	}
	return \@effects;
}

sub reactant_string {
	my ($self, $label) = @_;

	croak "Must pass a reaction label to reactant_string()\n" unless (defined $label);
	return $self->{reactant_string}{$label};
}

sub reaction_string {
	my ($self, $label) = @_;

	croak "Must pass a reaction label to reaction_string()\n" unless (defined $label);
	return $self->{reaction_string}{$label};
}

sub full_string {
	my ($self, $label) = @_;

	croak "Must pass a reaction label to full_string()\n" unless (defined $label);
	return $self->{full_string}{$label};
}

sub rate_string {
	my ($self, $label) = @_;

	croak "Must pass a reaction label to rate_string()\n" unless (defined $label);
	return $self->{rate_string}{$label};
}

sub reaction_number {
	my ($self, $label) = @_;

	croak "Must pass a reaction label to reaction_number()\n" unless (defined $label);
	return $self->{reaction_number}{$label};
}

# Return a list of the products from the given reaction
sub products {
	my ($self, $reaction) = @_;

	croak "$reaction: No such reaction" unless (defined $self->{reactions}{$reaction});
	my @products = keys %{$self->{reactions}{$reaction}{products}};
	return \@products; 
}

# Return a list of the reactants from the given reaction
sub reactants {
	my ($self, $reaction) = @_;

	croak "$reaction: No such reaction" unless (defined $self->{reactions}{$reaction});
	my @reactants = keys %{$self->{reactions}{$reaction}{reactants}};
	return \@reactants;
}

sub all_reactions {
	my ($self) = @_;
	return [ sort keys %{ $self->{reactions} } ];
}

sub all_species {
	my ($self) = @_;
	my %species;
	$species{$_} = 1 for (keys %{ $self->{produced} }, keys %{ $self->{consumed} });
	return [ keys %species ];
}

sub family {
	my ($self, $options) = @_;
	croak("Must specify options to create new family") unless (defined $options);
	croak("Must specify family name") unless (defined $options->{'name'});
	my $family = $options->{'name'};
	if (defined $self->{produced}{$family} or defined $self->{consumed}{$family}) {
		croak("$family already defined as family or species");
	}
	croak("Must specify family members") unless (defined $options->{'members'});
	my $members = $options->{members};
	my $weights = {};
	if (defined $options->{'weights'}) {
		$weights = $options->{'weights'};
	}

    my (%reactions, %yields);
    foreach my $species (@$members) {
        my $L = $self->consuming($species); 
        my $P = $self->producing($species); 
        $reactions{$_} = 1 for (@$P, @$L); 
		my @reactions = keys %reactions;
        my $effects = $self->effect_on($species, \@reactions);
        if (defined $weights->{$species}) {
            $effects->[$_] *= $weights->{$species} for (0..$#reactions);
        }
        $yields{$reactions[$_]}{$species} += $effects->[$_] for (0..$#reactions);
    }

    my %family_yield;
    foreach my $reaction (keys %reactions) {
        $family_yield{$reaction} = 0;
        foreach my $species (@$members) {
            $family_yield{$reaction} += $yields{$reaction}{$species} if (exists $yields{$reaction}{$species});
			#print "$family $reaction $species $yields{$reaction}{$species}\n" if (exists $yields{$reaction}{$species});
        }
		#delete $family_yield{$reaction} if ($family_yield{$reaction} == 0);
        delete $family_yield{$reaction} if (abs $family_yield{$reaction} < 1e-10);
    }

    foreach my $reaction (keys %family_yield) {
        if ($family_yield{$reaction} > 0) {
			push @{ $self->{produced}{$family}{reactions} }, $reaction;
			$self->{reactions}{$reaction}{family_products}{$family} = $family_yield{$reaction};
			#print "$family produced yield $reaction $family_yield{$reaction}\n";
        } else {
			push @{ $self->{consumed}{$family}{reactions} }, $reaction;
			$self->{reactions}{$reaction}{family_reactants}{$family} = -$family_yield{$reaction};
			#print "$family consumed yield $reaction $family_yield{$reaction}\n";
        }
    }
}

1;
