#!/usr/bin/perl -w
#
#    add_tags.nox.pl: add the specified tags to the CESM source code for NOx or VOC
#                     tagging, based on template source code files
#    Copyright (C) 2017  Tim Butler <tim.butler@iass-potsdam.de>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#


use warnings;
use strict;

# Choose NOX or VOC tagging
my $what_to_tag = 'NOX';
die "Can only tag NOX or VOC" unless $what_to_tag eq 'NOX' or $what_to_tag eq 'VOC';

my $model_version = "cesm1_2_2_CAM";

# These are the tags to add
my @uctags = qw(STR LGT AIR BIO BMB ANT INI); # All tags which appply to precursor species
my @uc_x_tags = (@uctags, qw(XTR)); # All tags which only apply to Ox species

# Make a list of these in lowercase, too
my @lctags = @uctags;
my @lc_x_tags = @uc_x_tags;
@lctags = map { $_ =~ tr/A-Z/a-z/; $_ } @lctags;
@lc_x_tags = map { $_ =~ tr/A-Z/a-z/; $_ } @lc_x_tags;

# Get a list of all of the files to which we must add tags
my $template_dir = "./templates/$model_version";
opendir DIR, $template_dir or die $!;
my @template_sources = grep /src/, readdir DIR;
closedir DIR;
die "No template source directories found" unless @template_sources;
my @template_files = ();
foreach my $dir (@template_sources) {
	opendir DIR, "$template_dir/$dir" or die $!;
	push @template_files, map { "$dir/$_" } grep /F90$/, readdir DIR;
	closedir DIR;
}
die "No template files found" unless @template_files;

# Make sure the SourceMods structure is in place
my $sourcemods_dir = "./SourceMods";
if (! -d $sourcemods_dir) {
	mkdir $sourcemods_dir or die "Unable to create $sourcemods_dir: $!\n";
}
foreach my $dir (@template_sources) {
	if (! -d "$sourcemods_dir/$dir") {
		mkdir "$sourcemods_dir/$dir" or die "Unable to create $sourcemods_dir/$dir: $!\n";
	}
}

# Copy over each of the template files, adding the requested tags.
foreach my $file (@template_files) {
	open FILE, "$template_dir/$file" or die $!;
	my @lines = <FILE>;
	close FILE;
	open FILE, ">$sourcemods_dir/$file" or die $!;
	for (my $l = 0; $l < @lines; $l++) {
		my $line = $lines[$l];
		if ($line =~ /START $what_to_tag TAGGING CODE/) {
			my $template_text;
			for ($l++; $lines[$l] !~ /END $what_to_tag TAGGING CODE/; $l++) {
				$template_text .= $lines[$l];
			}
			$l--;
			$line .= replacement_text($template_text, \@uctags, \@lctags);
		} elsif ($line =~ /START TAGGING CODE/) {
			my $template_text;
			for ($l++; $lines[$l] !~ /END TAGGING CODE/; $l++) {
				$template_text .= $lines[$l];
			}
			$l--;
			$line .= replacement_text($template_text, \@uc_x_tags, \@lc_x_tags);
		}
		print FILE $line;
	}
	close FILE;
}

sub replacement_text {
	my ($template_text, $uctags, $lctags) = @_;
	my $replacement_text = "";
	foreach my $t (0..$#$uctags) {
		my $tmp = $template_text;
		$tmp =~ s/_TAG/_$uctags->[$t]/sg;
		$tmp =~ s/_tag/_$lctags->[$t]/sg;
		$replacement_text .= $tmp;
	}
	return $replacement_text;
}
